# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module wrapping around the JMS root endpoints."""
import json
import logging
import os
from typing import Dict, List, Union
import uuid

import backoff
import requests

from ansys.hps.client.client import Client
from ansys.hps.client.common import Object
from ansys.hps.client.exceptions import HPSError
from ansys.hps.client.jms.resource import Operation, Permission, Project, TaskDefinitionTemplate
from ansys.hps.client.jms.schema.project import ProjectSchema

from .base import copy_objects as base_copy_objects
from .base import create_objects, delete_objects, get_object, get_objects, update_objects

log = logging.getLogger(__name__)


class JmsApi(object):
    """Wraps around the JMS root endpoints.

    Parameters
    ----------
    client : Client
        HPS client object.

    Examples
    --------

    Create a project.

    >>> from ansys.hps.client import Client
    >>> from ansys.hps.client.jms import JmsApi, Project
    >>> cl = Client(
    ...     url="https://127.0.0.1:8443/hps", username="repuser", password="repuser"
    ... )
    >>> jms_api = JmsApi(cl)
    >>> project = jms_api.create_project(Project(name="Example project"))

    """

    def __init__(self, client: Client):
        """Initialize JMS API."""
        self.client = client
        self._fs_url = None

    @property
    def url(self) -> str:
        """URL of the API."""
        return f"{self.client.url}/jms/api/v1"

    @property
    def fs_url(self) -> str:
        """URL of the file storage gateway."""
        if self._fs_url is None:
            self._fs_url = _find_available_fs_url(self.get_storage())
        return self._fs_url

    def get_api_info(self):
        """Get information of the JMS API that the client is connected to.

        Information includes the version and build date.
        """
        r = self.client.session.get(self.url)
        return r.json()

    ################################################################
    # Projects
    def get_projects(self, as_objects=True, **query_params) -> List[Project]:
        """Get a list of projects, optionally filtered by query parameters."""
        return get_projects(self.client, self.url, as_objects, **query_params)

    def get_project(self, id: str) -> Project:
        """Get a single project for a given project ID."""
        return get_project(self.client, self.url, id)

    def get_project_by_name(
        self, name: str, last_created: bool = True
    ) -> Union[Project, List[Project]]:
        """Query projects by name.

        If no projects are found, an empty list is returned.

        In multiple projects with same name are found, what is returned depends
        on the ``last_created`` value:

        - If ``last_created=True``, the last created project is returned.
        - If ``last_created=False``, the full list of projects with the given name is returned.

        """
        return get_project_by_name(self.client, self.url, name, last_created)

    def create_project(self, project: Project, replace=False, as_objects=True) -> Project:
        """Create a project."""
        return create_project(self.client, self.url, project, replace, as_objects)

    def update_project(self, project: Project, as_objects=True) -> Project:
        """Update a project."""
        return update_project(self.client, self.url, project, as_objects)

    def delete_project(self, project):
        """Delete a project."""
        return delete_project(self.client, self.url, project)

    def restore_project(self, path: str) -> Project:
        """Restore a project from an archive.

        Parameters
        ----------
        path : str
            Path of the archive file.

        """
        return _restore_project(self, path)

    ################################################################
    # Task Definition Templates
    def get_task_definition_templates(
        self, as_objects=True, **query_params
    ) -> List[TaskDefinitionTemplate]:
        """Get a list of task definition templates, optionally filtered by query parameters."""
        return get_objects(
            self.client.session, self.url, TaskDefinitionTemplate, as_objects, **query_params
        )

    def create_task_definition_templates(
        self, templates: List[TaskDefinitionTemplate], as_objects=True, **query_params
    ) -> List[TaskDefinitionTemplate]:
        """Create task definition templates.

        Parameters
        ----------
        templates : list of :class:`ansys.hps.client.jms.TaskDefinitionTemplate`
            List of task definition templates.
        """
        return create_objects(
            self.client.session,
            self.url,
            templates,
            TaskDefinitionTemplate,
            as_objects,
            **query_params,
        )

    def update_task_definition_templates(
        self, templates: List[TaskDefinitionTemplate], as_objects=True, **query_params
    ) -> List[TaskDefinitionTemplate]:
        """Update task definition templates.

        Parameters
        ----------
        templates : list of :class:`ansys.hps.client.jms.TaskDefinitionTemplate`
            List of task definition templates.
        """
        return update_objects(
            self.client.session,
            self.url,
            templates,
            TaskDefinitionTemplate,
            as_objects,
            *query_params,
        )

    def delete_task_definition_templates(self, templates: List[TaskDefinitionTemplate]):
        """Delete task definition templates.

        Parameters
        ----------
        templates : list of :class:`ansys.hps.client.jms.TaskDefinitionTemplate`
            List of task definition templates.
        """
        return delete_objects(self.client.session, self.url, templates, TaskDefinitionTemplate)

    def copy_task_definition_templates(
        self, templates: List[TaskDefinitionTemplate], wait: bool = True
    ) -> Union[str, List[str]]:
        """Create task definition templates by copying existing templates.

        Parameters
        ----------
        templates : List[TaskDefinitionTemplate]
            List of task definition template. Note that only the ``id`` field of
            ``TaskDefinitionTemplate`` objects must be filled. The other fields can be empty.
        wait : bool, optional
            Whether to wait for the copy to complete. The default is ``True``.

        Returns
        -------
        Union[List[str], str]
            If ``wait=True``, returns the list of newly created template IDs.
            If ``wait=False``, returns an operation ID that can be used to
            track progress.
        """
        return _copy_objects(self.client, self.url, templates, wait=wait)

    # Task Definition Template Permissions
    def get_task_definition_template_permissions(
        self, template_id: str, as_objects: bool = True
    ) -> List[Permission]:
        """Get permissions of a task definition template."""
        return get_objects(
            self.client.session,
            f"{self.url}/task_definition_templates/{template_id}",
            Permission,
            as_objects,
        )

    def update_task_definition_template_permissions(
        self,
        template_id: str,
        permissions: List[Permission],
        as_objects: bool = True,
    ) -> List[Permission]:
        """Update permissions of a task definition template."""
        return update_objects(
            self.client.session,
            f"{self.url}/task_definition_templates/{template_id}",
            permissions,
            Permission,
            as_objects,
        )

    ################################################################
    # Operations
    def get_operations(self, as_objects=True, **query_params) -> List[Operation]:
        """Get operations."""
        return get_objects(
            self.client.session, self.url, Operation, as_objects=as_objects, **query_params
        )

    def get_operation(self, id, as_object=True) -> Operation:
        """Get an operation."""
        return get_object(self.client.session, self.url, Operation, id, as_object=as_object)

    def monitor_operation(self, operation_id: str, max_value: float = 5.0, max_time: float = None):
        """Poll an operation until it is completed using an exponential backoff.

        Parameters
        ----------
        operation_id : str
            ID of the operation to monitor.
        max_value: float, optional
            Maximum interval in seconds between consecutive calls.
        max_time: float, optional
            Maximum time in seconds to pool the operation before giving up.
        """
        return _monitor_operation(self, operation_id, max_value, max_time)

    ################################################################
    # Storages
    def get_storage(self):
        """Get a list of storages."""
        return _get_storages(self.client, self.url)


def get_projects(client, api_url, as_objects=True, **query_params) -> List[Project]:
    """
    Get a list of projects.
    """
    url = f"{api_url}/projects"
    r = client.session.get(url, params=query_params)

    data = r.json()["projects"]
    if not as_objects:
        return data

    schema = ProjectSchema(many=True)
    return schema.load(data)


def get_project(client, api_url, id) -> Project:
    """
    Get a single project.
    """

    url = f"{api_url}/projects/{id}"
    r = client.session.get(url)

    if len(r.json()["projects"]):
        schema = ProjectSchema()
        return schema.load(r.json()["projects"][0])
    return None


def get_project_by_name(client, api_url, name, last_created=True) -> Union[Project, List[Project]]:
    """
    Get a single project by name.
    """

    params = {"name": name}
    if last_created:
        params["sort"] = "-creation_time"
        params["limit"] = 1

    projects = get_projects(client, api_url, **params)

    if len(projects) == 1:
        return projects[0]
    return projects


def create_project(client, api_url, project, replace=False, as_objects=True) -> Project:
    """Create a project."""
    url = f"{api_url}/projects/"

    schema = ProjectSchema()
    serialized_data = schema.dump(project)
    json_data = json.dumps({"projects": [serialized_data], "replace": replace})
    r = client.session.post(f"{url}", data=json_data)

    if not r.json()["projects"]:
        raise HPSError(f"Failed to create the project. Request response: {r.json()}")

    data = r.json()["projects"][0]
    if not as_objects:
        return data

    return schema.load(data)


def update_project(client, api_url, project, as_objects=True) -> Project:
    """Update a project."""
    url = f"{api_url}/projects/{project.id}"

    schema = ProjectSchema()
    serialized_data = schema.dump(project)
    json_data = json.dumps({"projects": [serialized_data]})
    r = client.session.put(f"{url}", data=json_data)

    data = r.json()["projects"][0]
    if not as_objects:
        return data

    return schema.load(data)


def delete_project(client, api_url, project):
    """Delete a project."""
    url = f"{api_url}/projects/{project.id}"
    r = client.session.delete(url)


def _monitor_operation(
    jms_api: JmsApi, operation_id: str, max_value: float = 5.0, max_time: float = None
) -> Operation:
    """Monitor an operation."""

    @backoff.on_predicate(
        backoff.expo,
        lambda x: x[1] == False,
        jitter=backoff.full_jitter,
        max_value=max_value,
        max_time=max_time,
    )
    def _monitor():
        """Monitor the operation with its ID."""
        done = False
        op = jms_api.get_operation(id=operation_id)
        if op:
            done = op.finished
        return op, done

    op, done = _monitor()

    if not done:
        raise HPSError(f"Operation {operation_id} did not complete.")
    return op


def _copy_objects(
    client: Client, api_url: str, objects: List[Object], wait: bool = True
) -> Union[str, List[str]]:
    """Copy objects."""
    operation_id = base_copy_objects(client.session, api_url, objects)

    if not wait:
        return operation_id

    op = _monitor_operation(JmsApi(client), operation_id, 1.0)
    if not op.succeeded:
        obj_type = objects[0].__class__
        rest_name = obj_type.Meta.rest_name
        raise HPSError(f"Failed to copy {rest_name} with ids = {[obj.id for obj in objects]}.")
    return op.result["destination_ids"]


def _restore_project(jms_api, archive_path):
    """Restore an archived project."""
    if not os.path.exists(archive_path):
        raise HPSError(f"Project archive: path does not exist {archive_path}")

    # Upload archive to FS API
    archive_name = os.path.basename(archive_path)

    bucket = f"hps-client-restore-{uuid.uuid4()}"
    fs_file_url = f"{jms_api.client.url}/fs/api/v1/{bucket}/{archive_name}"
    ansfs_file_url = f"ansfs://{bucket}/{archive_name}"  # noqa: E231

    fs_headers = {"content-type": "application/octet-stream"}

    log.info(f"Uploading archive to {fs_file_url}")
    with open(archive_path, "rb") as file_content:
        r = jms_api.client.session.post(fs_file_url, data=file_content, headers=fs_headers)

    # POST restore request
    log.info(f"Restoring archive from {ansfs_file_url}")
    url = f"{jms_api.url}/projects/archive"
    query_params = {"backend_path": ansfs_file_url}
    r = jms_api.client.session.post(url, params=query_params)

    # Monitor restore operation
    operation_location = r.headers["location"]
    log.debug(f"Operation location: {operation_location}")
    operation_id = operation_location.rsplit("/", 1)[-1]
    log.debug(f"Operation id: {operation_id}")

    op = jms_api.monitor_operation(operation_id)

    if not op.succeeded:
        raise HPSError(f"Failed to restore project from archive {archive_path}.")

    project_id = op.result["project_id"]
    log.info(f"Done restoring project, project_id = '{project_id}'")

    # Delete archive file on server
    log.info(f"Delete temporary bucket {bucket}")
    r = jms_api.client.session.put(f"{jms_api.client.url}/fs/api/v1/remove/{bucket}")

    return get_project(jms_api.client, jms_api.url, project_id)


def _get_storages(client: Client, api_url: str) -> List[Dict]:
    """
    Get a list of storages.
    """
    url = f"{api_url}/storage"
    r = client.session.get(url)
    return r.json()["backends"]


def _find_available_fs_url(file_storages: Dict) -> str:
    """Find first available file storage URL."""

    if not file_storages:
        raise HPSError("There is no file storage information.")

    rest_gateways = [fs for fs in file_storages if fs["obj_type"] == "RestGateway"]
    rest_gateways.sort(key=lambda fs: fs["priority"])

    if not rest_gateways:
        raise HPSError("There is no file storage gateway defined.")

    for d in rest_gateways:
        url = d["url"]
        try:
            r = requests.get(url, verify=False, timeout=2)
            is_ansft = r.json()["ansft"]
        except Exception as ex:
            log.debug(ex)
            continue
        if r.status_code == 200 and is_ansft:
            return url

    raise HPSError(
        f"All defined file storage gateways are unavailable"
        f" ({', '.join([d['url'] for d in rest_gateways])})."
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing float, integer, boolean, and string parameter definition schema."""
import logging

from marshmallow import fields
from marshmallow_oneofschema import OneOfSchema

from ansys.hps.client.common import ObjectSchemaWithModificationInfo

log = logging.getLogger(__name__)


class ParameterDefinitionBaseSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    name = fields.String(allow_none=True, metadata={"description": "Name (ID) of the parameter."})

    quantity_name = fields.String(
        allow_none=True,
        metadata={
            "description": "Name of the quantity that the parameter represents. "
            "For example, ``Length``."
        },
    )
    units = fields.String(allow_none=True, metadata={"description": "Units for the parameter."})
    display_text = fields.String(
        allow_none=True, metadata={"description": "Text to display as the parameter name."}
    )

    mode = fields.String(
        load_only=True,
        metadata={
            "description": "Indicates whether it's an input "
            "or output parameter. The mode is filled server-side."
        },
    )


class FloatParameterDefinitionSchema(ParameterDefinitionBaseSchema):
    class Meta(ParameterDefinitionBaseSchema.Meta):
        pass

    type = fields.Constant("float")
    default = fields.Float(allow_none=True, metadata={"description": "Default parameter value."})
    lower_limit = fields.Float(
        allow_none=True, metadata={"description": "Lower bound for the parameter value."}
    )
    upper_limit = fields.Float(
        allow_none=True, metadata={"description": "Upper bound for the parameter value."}
    )
    step = fields.Float(
        allow_none=True,
        metadata={
            "description": "If provided, allowable values are given by: "
            "AllowableValue = lower_limit + n * step, "
            "where n is an integer and AllowableValue <= upper_limit."
        },
    )
    cyclic = fields.Bool(
        allow_none=True, metadata={"description": "Whether the parameter is cyclic."}
    )
    value_list = fields.List(
        fields.Float(),
        allow_none=True,
        metadata={
            "description": "List of allowed values. This parameter provides an alternative "
            "to specifying upper and lower limits."
        },
    )


class IntParameterDefinitionSchema(ParameterDefinitionBaseSchema):
    class Meta(ParameterDefinitionBaseSchema.Meta):
        pass

    type = fields.Constant("int")
    default = fields.Integer(allow_none=True, metadata={"description": "Default parameter value."})
    lower_limit = fields.Integer(
        allow_none=True, metadata={"description": "Lower bound for the parameter value."}
    )
    upper_limit = fields.Integer(
        allow_none=True, metadata={"description": "Upper bound for the parameter value."}
    )
    step = fields.Integer(allow_none=True, metadata={"description": "The default is ``1``."})
    cyclic = fields.Bool(
        allow_none=True, metadata={"description": "Whether the parameter is cyclic."}
    )


class BoolParameterDefinitionSchema(ParameterDefinitionBaseSchema):
    class Meta(ParameterDefinitionBaseSchema.Meta):
        pass

    type = fields.Constant("bool")
    default = fields.Bool(allow_none=True, metadata={"description": "Default parameter value."})


class StringParameterDefinitionSchema(ParameterDefinitionBaseSchema):
    class Meta(ParameterDefinitionBaseSchema.Meta):
        pass

    type = fields.Constant("string")
    default = fields.String(allow_none=True, metadata={"description": "Default parameter value."})
    value_list = fields.List(
        fields.String(), allow_none=True, metadata={"description": "List of allowed values."}
    )


class ParameterDefinitionSchema(OneOfSchema):
    type_field = "type"

    type_schemas = {
        "float": FloatParameterDefinitionSchema,
        "int": IntParameterDefinitionSchema,
        "bool": BoolParameterDefinitionSchema,
        "string": StringParameterDefinitionSchema,
    }

    def get_obj_type(self, obj):
        return obj.__class__.__name__.replace("ParameterDefinition", "").lower()



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing task schema."""

from marshmallow import fields

from ansys.hps.client.common import ObjectSchemaWithModificationInfo

from .object_reference import IdReference, IdReferenceList
from .task_definition import TaskDefinitionSchema


class TaskSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    pending_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={"description": "Date and time that the task was set to pending."},
    )
    prolog_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={"description": "Date and time that the task was set to prolog."},
    )
    running_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={"description": "Date and time that the task was set to running."},
    )
    finished_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={"description": "Date and time that the task was completed."},
    )

    eval_status = fields.String(allow_none=True, metadata={"description": "Evaluation status."})
    trial_number = fields.Integer(
        allow_none=True,
        load_only=True,
        metadata={"description": "Which attempt to execute the process step this task represents."},
    )
    elapsed_time = fields.Float(
        allow_none=True,
        load_only=True,
        metadata={"description": "Number of seconds it took the evaluator to execute the task."},
    )

    task_definition_id = IdReference(
        allow_none=False,
        attribute="task_definition_id",
        referenced_class="TaskDefinition",
        metadata={
            "description": "ID of the :class:`TaskDefinition` instance that the task is linked to."
        },
    )
    task_definition_snapshot = fields.Nested(
        TaskDefinitionSchema,
        allow_none=True,
        metadata={
            "description": "Snapshot of the  :class:`TaskDefinition` instance that was created"
            " when the task status changed to prolog, before evaluation."
        },
    )

    executed_command = fields.String(allow_none=True)

    job_id = IdReference(
        allow_none=False,
        attribute="job_id",
        referenced_class="Job",
        metadata={"description": "ID of the :class:`Job` instance that the task is linked to."},
    )

    host_id = fields.String(
        allow_none=True,
        metadata={"description": "UUID of the :class:`Evaluator` instance that updated the task."},
    )

    input_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="input_file_ids",
        metadata={"description": "List of IDs of input files of the task."},
    )
    output_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="output_file_ids",
        metadata={"description": "List of IDs of output files of the task."},
    )
    monitored_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="monitored_file_ids",
        metadata={"description": "List of IDs of monitored files of the task."},
    )

    inherited_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="inherited_file_ids",
        metadata={"description": "List of IDs of inherited files of the task."},
    )
    owned_file_ids = IdReferenceList(
        referenced_class="File",
        attribute="owned_file_ids",
        metadata={"description": "List of IDs of owned files of the task."},
    )

    license_context_id = fields.String(
        allow_none=True,
        metadata={"description": "ID of the license context in use."},
    )

    custom_data = fields.Dict(
        allow_none=True,
        metadata={"description": "Dictionary type field for storing custom data."},
    )

    working_directory = fields.String(
        allow_none=True,
        metadata={"description": "Working directory of the task."},
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing the file resource."""
from datetime import datetime
import io
from typing import Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.file import FileSchema


class File(Object):
    """Provides the file resource.

    Parameters
    ----------
    src : Union[str, io.IOBase],  optional
        Client-only field to specify either the path of an input file
        or a file-like object. In the latter case, `requests` recommends that
        you open files in binary mode.
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    name : str
        Name of the file resource.
    type : str, optional
        Type of the file. While you can use any string, you should use the correct media type
        for the given resource.
    storage_id : str, optional
        ID of the file in the (orthogonal) file storage system.
    size : int, optional
    hash : str, optional
    creation_time : datetime, optional
        Date and time that the file resource was created.
    modification_time : datetime, optional
        Date and time that the file resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    format : str, optional
    expiry_time : datetime, optional
        File expiration time.
    evaluation_path : str, optional
        Relative path for the file instance to store the job evaluation under.
    monitor : bool, optional
        Whether to live monitor the file's content.
    collect : bool, optional
        Whether the file should be collected per job.
    collect_interval : int, optional
        Collection frequency for a file with ``collect=True``. The minimum value
        is limited by the evaluator's settings. A value of ``0`` or ``None`` indicates
        that the evaluator should decide. Another value indicates the interval in seconds.
    reference_id : str, optional
        ID of the reference file that this file was created from.
    """

    class Meta:
        schema = FileSchema
        rest_name = "files"

    def __init__(
        self,
        src: Union[str, io.IOBase] = None,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        type: str = missing,
        storage_id: str = missing,
        size: int = missing,
        hash: str = missing,
        expiry_time: datetime = missing,
        format: str = missing,
        evaluation_path: str = missing,
        monitor: bool = missing,
        collect: bool = missing,
        collect_interval: int = missing,
        reference_id: str = missing,
        **kwargs,
    ):
        self.src = src
        self.content = None

        self.id = id
        self.name = name
        self.type = type
        self.storage_id = storage_id
        self.size = size
        self.hash = hash
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.expiry_time = expiry_time
        self.format = format
        self.evaluation_path = evaluation_path
        self.monitor = monitor
        self.collect = collect
        self.collect_interval = collect_interval
        self.reference_id = reference_id

        self.obj_type = self.__class__.__name__


FileSchema.Meta.object_class = File



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the job resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.job import JobSchema


class Job(Object):
    """Provides the job resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name of the job.
    eval_status : str
        Evaluation status.
    job_definition_id : str
        ID of the linked job definition. For more information, see the :class:`JobDefinition` class.
    priority : int, optional
        Priority for evaluating the job. The default is ``0``, which is the highest priority. Assigning a higher value to a job makes it a lower priority.
    values : dict[str, any], optional
        Dictionary with (name,value) pairs for all parameters defined in the linked job definition.
    fitness : float, optional
        Fitness value computed.
    fitness_term_values : dict[str, float], optional
        Dictionary with (name,value) pairs for all fitness terms computed.
    note : str, optional
        Note for the job.
    creator : str, optional
        Additional information about the creator of the job.
    executed_level : int, optional
        Execution level of the last executed task. A value of ``-1`` indicates that no task has been executed yet.
    elapsed_time : float
        Number of seconds it took the evaluators to update the job.
    host_ids : list, optional
        List of host IDs of the evaluators that updated the job.
    file_ids : list[str]
        List of IDs of all files of the job.
    """

    class Meta:
        schema = JobSchema
        rest_name = "jobs"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        eval_status: str = missing,
        job_definition_id: str = missing,
        priority: int = missing,
        values: Dict[str, any] = missing,
        fitness: float = missing,
        fitness_term_values: Dict[str, float] = missing,
        note: str = missing,
        creator: str = missing,
        executed_level: int = missing,
        elapsed_time: float = missing,
        host_ids: List = missing,
        file_ids: List[str] = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.eval_status = eval_status
        self.job_definition_id = job_definition_id
        self.priority = priority
        self.values = values
        self.fitness = fitness
        self.fitness_term_values = fitness_term_values
        self.note = note
        self.creator = creator
        self.executed_level = executed_level
        self.elapsed_time = elapsed_time
        self.host_ids = host_ids
        self.file_ids = file_ids

        self.obj_type = self.__class__.__name__


JobSchema.Meta.object_class = Job



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module wrapping around RMS root endpoints."""
import logging
from typing import List

from ansys.hps.client.client import Client
from ansys.hps.client.rms.models import (
    ClusterInfo,
    ComputeResourceSet,
    EvaluatorConfiguration,
    EvaluatorConfigurationUpdate,
    EvaluatorRegistration,
    ScalerRegistration,
)

from .base import create_objects, get_object, get_objects, get_objects_count

log = logging.getLogger(__name__)


class RmsApi(object):
    """Wraps around the RMS root endpoints.

    Parameters
    ----------
    client : Client
        HPS client object.
    """

    def __init__(self, client: Client):
        self.client = client
        self._fs_url = None

    @property
    def url(self) -> str:
        """URL of the API."""
        return f"{self.client.url}/rms/api/v1"

    def get_api_info(self):
        """Get information on the RMS API the client is connected to.

        The information includes the version and build date.
        """
        r = self.client.session.get(self.url)
        return r.json()

    ################################################################
    # Evaluators
    def get_evaluators_count(self, **query_params) -> int:
        """Get the number of evaluators, optionally filtered by query parameters."""
        return get_objects_count(
            self.client.session, self.url, EvaluatorRegistration, **query_params
        )

    def get_evaluators(self, as_objects=True, **query_params) -> List[EvaluatorRegistration]:
        """Get a list of evaluators, optionally filtered by query parameters.

        The server only returns the first 10 objects (``limit=10``) by default.
        """
        return get_objects(
            self.client.session, self.url, EvaluatorRegistration, as_objects, **query_params
        )

    def get_evaluator_configuration(self, id: str, as_object=True) -> EvaluatorConfiguration:
        """Get an evaluator's configuration"""
        return get_object(
            self.client.session,
            f"{self.url}/evaluators/{id}/configuration",
            EvaluatorConfiguration,
            id=id,
            as_object=as_object,
        )

    def update_evaluator_configuration(
        self, id: str, configuration: EvaluatorConfigurationUpdate, as_object=True
    ) -> EvaluatorConfigurationUpdate:
        """Update an evaluator configuration.

        Examples
        --------

        Set a custom resource property on a Linux evaluator that was active in the past 60 seconds.

        >>> import datetime
        >>> from ansys.hps.client import Client
        >>> from ansys.hps.client.jms import RmsApi, EvaluatorConfigurationUpdate
        >>> cl = Client(
        ...     url="https://localhost:8443/hps", username="repuser", password="repuser"
        ... )
        >>> rms_api = RmsApi(cl)
        >>> query_params = {
        ...     "platform" : "linux",
        ...     "update_time.gt" : datetime.datetime.utcnow() - datetime.timedelta(seconds=60)
        ... }
        >>> evaluator = rms_api.get_evaluators(fields=["id", "host_id"], **query_params)[0]
        >>> config_update = EvaluatorConfigurationUpdate(
        ...    custom_resource_properties={"disk_type" : "SSD"}
        ... )
        >>> rms_api.update_evaluator_configuration(ev.id, config_update)

        """
        return create_objects(
            self.client.session, f"{self.url}/evaluators/{id}", [configuration], as_object
        )[0]

    ################################################################
    # Scalers
    def get_scalers_count(self, **query_params) -> int:
        """Get the number of scalers, optionally filtered by query parameters."""
        return get_objects_count(self.client.session, self.url, ScalerRegistration, **query_params)

    def get_scalers(self, as_objects=True, **query_params) -> List[ScalerRegistration]:
        """Get a list of scalers, optionally filtered by query parameters.

        The server only returns the first 10 objects (``limit=10``) by default.
        """
        return get_objects(
            self.client.session, self.url, ScalerRegistration, as_objects, **query_params
        )

    ################################################################
    # Compute resource sets
    def get_compute_resource_sets_count(self, **query_params) -> int:
        """Get the number of compute resource sets, optionally filtered by query parameters."""
        return get_objects_count(self.client.session, self.url, ComputeResourceSet, **query_params)

    def get_compute_resource_sets(
        self, as_objects=True, **query_params
    ) -> List[ComputeResourceSet]:
        """Get a list of compute resource sets, optionally filtered by query parameters.

        The server only returns the first 10 objects (``limit=10``) by default.
        """
        return get_objects(
            self.client.session, self.url, ComputeResourceSet, as_objects, **query_params
        )

    def get_compute_resource_set(self, id, as_object=True) -> ComputeResourceSet:
        """Get a compute resource set."""
        return get_object(
            self.client.session,
            f"{self.url}/compute_resource_sets/{id}",
            ComputeResourceSet,
            as_object,
            from_collection=True,
        )

    def get_cluster_info(self, compute_resource_set_id, as_object=True) -> ClusterInfo:
        """Get the cluster information of a compute resource set."""

        return get_object(
            self.client.session,
            f"{self.url}/compute_resource_sets/{compute_resource_set_id}/cluster_info",
            ClusterInfo,
            as_object=as_object,
        )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the job definition resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.job_definition import JobDefinitionSchema
from .fitness_definition import FitnessDefinition


class JobDefinition(Object):
    """Provides the job definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name of the job definition.
    active : bool
        Whether this job definition is active.
    client_hash : str, optional
    parameter_definition_ids : list[str]
        List of parameter definition IDs.
    parameter_mapping_ids : list[str]
        List of parameter mapping IDs.
    task_definition_ids : list[str]
        List of task definition IDs.
    fitness_definition : FitnessDefinition, optional
        A :class:`FitnessDefinition` object.
    """

    class Meta:
        schema = JobDefinitionSchema
        rest_name = "job_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        active: bool = missing,
        client_hash: str = missing,
        parameter_definition_ids: List[str] = missing,
        parameter_mapping_ids: List[str] = missing,
        task_definition_ids: List[str] = missing,
        fitness_definition: FitnessDefinition = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.active = active
        self.client_hash = client_hash
        self.parameter_definition_ids = parameter_definition_ids
        self.parameter_mapping_ids = parameter_mapping_ids
        self.task_definition_ids = task_definition_ids
        self.fitness_definition = fitness_definition

        self.obj_type = self.__class__.__name__


JobDefinitionSchema.Meta.object_class = JobDefinition



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyHPS is a Python client for Ansys HPC Platform Services (HPS)."""

from .__version__ import __ansys_apps_version__, __version__
from .auth import AuthApi
from .authenticate import authenticate
from .client import Client
from .exceptions import APIError, ClientError, HPSError
from .jms import JmsApi, ProjectApi
from .rms import RmsApi
from .warnings import UnverifiedHTTPSRequestsWarning



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing the Python interface to the Authorization Service API."""

from typing import Dict, List

from ansys.hps.client.client import Client

from ..resource import User
from ..schema.user import UserSchema


class AuthApi:
    """Provides a minimal wrapper around the Keycloak API to query user information.

    Parameters
    ----------
    client : Client
        HPS client object.

    Examples
    --------

    Get users whose first name contains ``john``.

    >>> from ansys.hps.client import Client
    >>> from ansys.hps.client.auth import AuthApi, User
    >>> cl = Client(
    ...     url="https://127.0.0.1:8443/hps", username="repuser", password="repuser"
    ... )
    >>> auth_api = AuthApi(cl)
    >>> users = auth_api.get_users(firstName="john", exact=False)

    """

    def __init__(self, client: Client):
        self.client = client

    @property
    def url(self) -> str:
        """API URL."""
        return f"{self.client.url}/auth"

    @property
    def realm_url(self) -> str:
        """Realm URL."""
        return f"{self.url}/admin/realms/{self.client.realm}"

    def get_users(self, as_objects=True, **query_params) -> List[User]:
        """Get users, filtered according to query parameters.

        Examples of query parameters are:

        - ``username``
        - ``firstName``
        - ``lastName``
        - ``exact``

        Pagination is also supported using the ``first`` and ``max`` parameters.

        For a list of supported query parameters, see the
        `Keycloak API documentation <https://www.keycloak.org/documentation>`__.

        """
        r = self.client.session.get(url=f"{self.realm_url}/users", params=query_params)
        data = r.json()

        if not as_objects:
            return data

        schema = UserSchema(many=True)
        return schema.load(data)

    def get_user(self, id: str, as_object: bool = True) -> User:
        """Get the user representation for a given user ID."""
        r = self.client.session.get(
            url=f"{self.realm_url}/users/{id}",
        )
        data = r.json()
        if not as_object:
            return data

        schema = UserSchema(many=False)
        return schema.load(data)

    def get_user_groups_names(self, id: str) -> List[str]:
        """Get the name of the groups that the user belongs to."""
        return [g["name"] for g in self.get_user_groups(id)]

    def get_user_realm_roles_names(self, id: str) -> List[str]:
        """Get the name of the realm roles for the user."""
        return [r["name"] for r in self.get_user_realm_roles(id)]

    def get_user_groups(self, id: str) -> List[Dict]:
        """Get the groups that the user belongs to."""
        r = self.client.session.get(
            url=f"{self.realm_url}/users/{id}/groups",
        )
        return r.json()

    def get_user_realm_roles(self, id: str) -> List[Dict]:
        """Get the realm roles for the user."""
        r = self.client.session.get(
            url=f"{self.realm_url}/users/{id}/role-mappings/realm",
        )
        return r.json()

    def user_is_admin(self, id: str) -> bool:
        """Determine if the user is a system administrator."""

        from ansys.hps.client.jms import JmsApi

        # the admin keys are configurable settings of JMS
        # they need to be queried, can't be hardcoded
        jms_api = JmsApi(self.client)
        admin_keys = jms_api.get_api_info()["settings"]["admin_keys"]

        # query user groups and roles and store in the same format
        # as admin keys
        user_keys = [f"groups.{name}" for name in self.get_user_groups_names(id)] + [
            f"roles.{name}" for name in self.get_user_realm_roles_names(id)
        ]

        # match admin and user keys
        if set(admin_keys).intersection(user_keys):
            return True

        return False



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing file schema."""

from marshmallow import fields

from ansys.hps.client.common import ObjectSchemaWithModificationInfo

from .object_reference import IdReference


class FileSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    name = fields.String(metadata={"description": "Name of the file resource."})
    type = fields.String(
        allow_none=True,
        metadata={
            "description": "Type of the file. This can be any string but using a correct media "
            "type for the given resource is advisable."
        },
    )
    storage_id = fields.String(
        allow_none=True,
        metadata={"description": "File's identifier in the (orthogonal) file storage system"},
    )

    size = fields.Int(allow_none=True)
    hash = fields.String(allow_none=True)

    expiry_time = fields.DateTime(
        allow_none=True,
        metadata={"description": "File expiration time."},
    )

    format = fields.String(allow_none=True)
    evaluation_path = fields.String(
        allow_none=True,
        metadata={
            "description": "Relative path under which the file instance for a "
            "job evaluation will be stored.",
        },
    )

    monitor = fields.Bool(
        allow_none=True, metadata={"description": "Whether to live monitor the file's content."}
    )
    collect = fields.Bool(
        allow_none=True, metadata={"description": "Whether file should be collected per job"}
    )
    collect_interval = fields.Int(
        allow_none=True,
        metadata={
            "description": "Collect frequency for a file with collect=True."
            " Min value limited by the evaluator's settings."
            " 0/None - let the evaluator decide,"
            " other value - interval in seconds"
        },
    )

    reference_id = IdReference(
        attribute="reference_id",
        referenced_class="File",
        allow_none=True,
        metadata={"description": "Reference file from which this one was created"},
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module creating, getting, updating, deleting and copying objects."""
import json
import logging
from typing import List, Type

from requests import Session

from ansys.hps.client.common import Object
from ansys.hps.client.exceptions import ClientError

log = logging.getLogger(__name__)


def get_objects(
    session: Session, url: str, obj_type: Type[Object], as_objects=True, **query_params
):
    """Get objects with a session, URL, and object type."""
    rest_name = obj_type.Meta.rest_name
    url = f"{url}/{rest_name}"
    r = session.get(url, params=query_params)

    if query_params.get("count"):
        return r.json()[f"num_{rest_name}"]

    data = r.json()[rest_name]
    if not as_objects:
        return data

    schema = obj_type.Meta.schema(many=True)
    return schema.load(data)


def get_object(
    session: Session, url: str, obj_type: Type[Object], id: str, as_object=True, **query_params
):
    """Get an object with a session, URL, object type, and object."""
    rest_name = obj_type.Meta.rest_name
    url = f"{url}/{rest_name}/{id}"
    r = session.get(url, params=query_params)

    data = r.json()[rest_name]
    if not as_object:
        return data

    schema = obj_type.Meta.schema(many=True)
    if len(data) == 0:
        return None
    elif len(data) == 1:
        return schema.load(data)[0]
    elif len(data) > 1:
        raise ClientError(
            f"Multiple {Object.__class__.__name__} objects with id={id}: {schema.load(data)}"
        )


def _check_object_types(objects: List[Object], obj_type: Type[Object]):
    """Check object types."""
    are_same = [isinstance(o, obj_type) for o in objects]
    if not all(are_same):
        actual_types = set([type(o) for o in objects])
        if len(actual_types) == 1:
            actual_types = actual_types.pop()
        raise ClientError(f"Wrong object types: expected '{obj_type}', got {actual_types}.")


def create_objects(
    session: Session,
    url: str,
    objects: List[Object],
    obj_type: Type[Object],
    as_objects=True,
    **query_params,
):
    """Create objects."""
    if not objects:
        return []

    _check_object_types(objects, obj_type)

    rest_name = obj_type.Meta.rest_name

    url = f"{url}/{rest_name}"
    schema = obj_type.Meta.schema(many=True)
    serialized_data = schema.dump(objects)
    json_data = json.dumps({rest_name: serialized_data})

    r = session.post(f"{url}", data=json_data, params=query_params)
    data = r.json()[rest_name]
    if not as_objects:
        return data

    return schema.load(data)


def update_objects(
    session: Session,
    url: str,
    objects: List[Object],
    obj_type: Type[Object],
    as_objects=True,
    **query_params,
):
    """Update objects."""
    if not objects:
        return []

    _check_object_types(objects, obj_type)

    rest_name = obj_type.Meta.rest_name

    url = f"{url}/{rest_name}"
    schema = obj_type.Meta.schema(many=True)
    serialized_data = schema.dump(objects)
    json_data = json.dumps({rest_name: serialized_data})
    r = session.put(f"{url}", data=json_data, params=query_params)

    data = r.json()[rest_name]
    if not as_objects:
        return data

    return schema.load(data)


def delete_objects(session: Session, url: str, objects: List[Object], obj_type: Type[Object]):
    """Delete objects."""
    if not objects:
        return

    _check_object_types(objects, obj_type)

    obj_type = objects[0].__class__
    rest_name = obj_type.Meta.rest_name
    url = f"{url}/{rest_name}"
    data = json.dumps({"source_ids": [obj.id for obj in objects]})

    r = session.delete(url, data=data)


def copy_objects(session: Session, url: str, objects: List[Object], wait: bool = True) -> str:
    """Copy objects."""
    are_same = [o.__class__ == objects[0].__class__ for o in objects[1:]]
    if not all(are_same):
        raise ClientError("Mixed object types")

    obj_type = objects[0].__class__
    rest_name = obj_type.Meta.rest_name
    url = f"{url}/{rest_name}:copy"  # noqa: E231

    source_ids = [obj.id for obj in objects]
    r = session.post(url, data=json.dumps({"source_ids": source_ids}))

    operation_location = r.headers["location"]
    operation_id = operation_location.rsplit("/", 1)[-1]

    return operation_id



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the project resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.project import ProjectSchema


class Project(Object):
    """Provides the project resource.

    Parameters
    ----------
    id : str
        Unique ID to access the project, assigned server side on creation.
    name : str
        Name of the project.
    active : bool
        Defines whether the project is active for evaluation.
    priority : int
        Priority for picking the project for evaluation.
    creation_time : datetime, optional
        Date and time that the project was created.
    modification_time : datetime, optional
        Date and time that the project was last modified.
    statistics : dict, optional
        Dictionary containing various project statistics.
    """

    class Meta:
        schema = ProjectSchema
        rest_name = "projects"

    def __init__(
        self,
        id: str = missing,
        name: str = missing,
        active: bool = missing,
        priority: int = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        statistics: Dict = missing,
        **kwargs,
    ):
        self.id = id
        self.name = name
        self.active = active
        self.priority = priority
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.statistics = statistics

        self.obj_type = self.__class__.__name__


ProjectSchema.Meta.object_class = Project



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module getting the :class:`requests.Session` object configured for HPS with a given access token
and pinging a given URL."""
import logging
from typing import Union

import requests
from requests.adapters import HTTPAdapter, Retry

log = logging.getLogger(__name__)


def create_session(
    access_token: str = None,
    verify: Union[bool, str] = True,
    disable_security_warnings=False,
) -> requests.Session:
    """Get the :class:`requests.Session` object configured for HPS with a given access token.

    Parameters
    ----------
    access_token : str
        Access token provided by the :meth:`ansys.hps.client.auth.authenticate` method.
    verify: Union[bool, str], optional
        If a Boolean, whether to verify the server's TLS certificate. The default is ``True``.
        If a string, the path to the CA bundle to use. For more information, see the
        :class:`requests.Session` documentation.
    disable_security_warnings: bool, optional
        Whether to disable warnings about insecure HTTPS requests. The default is ``False``.

    Returns
    -------
    :class:`requests.Session`
        Session object.
    """
    session = requests.Session()

    # Disable SSL certificate verification and warnings about it
    session.verify = verify

    if disable_security_warnings:
        requests.packages.urllib3.disable_warnings(
            requests.packages.urllib3.exceptions.InsecureRequestWarning
        )

    # Set basic content type to json
    session.headers = {
        "content-type": "application/json",
    }

    if access_token:
        session.headers.update({"Authorization": "Bearer %s" % access_token})

    retries = Retry(total=5, backoff_factor=0.5, status_forcelist=[502, 503, 504])
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session


def ping(session: requests.Session, url: str, timeout=10.0) -> bool:
    """Ping the given URL.

    Parameters
    ----------
    session : :class:`requests.Session`
        Session object.
    url : str
        URL address to ping.
    timeout : float, optional
        Time in seconds to continue pinging the URL before reporting a failure. The
        default is ``10.0``.

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.
    """
    log.debug("Ping %s ..." % url)
    r = session.get(url, timeout=timeout)
    success = r.status_code == requests.codes.ok
    if success:
        log.debug("Ping successful")
    else:
        log.debug("Ping failed, HTTP error %s" % r.status_code)
    return success



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module that creates the user schema."""

from marshmallow import fields

from ansys.hps.client.common.base_schema import BaseSchema


class UserSchema(BaseSchema):
    """Create user schema with ID, username, password, first name, last name, and email."""

    class Meta(BaseSchema.Meta):
        pass

    id = fields.String(
        load_only=True,
        metadata={"description": "Unique user ID, generated internally by the server on creation."},
    )
    username = fields.Str(metadata={"description": "Username."})
    password = fields.Str(dump_only=True, metadata={"description": "Password."})
    firstName = fields.Str(
        allow_none=True, metadata={"description": "First name."}, attribute="first_name"
    )
    lastName = fields.Str(
        allow_none=True, metadata={"description": "Last name."}, attribute="last_name"
    )
    email = fields.Str(allow_none=True, metadata={"description": "E-mail address."})



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the task definition template resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.task_definition_template import (
    TaskDefinitionTemplateSchema,
    TemplateInputFileSchema,
    TemplateOutputFileSchema,
    TemplatePropertySchema,
    TemplateResourceRequirementsSchema,
)
from .task_definition import HpcResources, Software


class TemplateProperty(Object):
    """Provides the template property resource.

    Parameters
    ----------
    default : any, optional
        Default value.
    description : str, optional
        Description of the property's purpose.
    type : str, optional
        Type of the property. Options are ``bool``, ``float``, ``int``, and ``string``.
    value_list : any, optional
        List of possible values for the property.
    """

    class Meta:
        schema = TemplatePropertySchema
        rest_name = "None"

    def __init__(
        self,
        default: Any = missing,
        description: str = missing,
        type: str = missing,
        value_list: Any = missing,
        **kwargs,
    ):
        self.default = default
        self.description = description
        self.type = type
        self.value_list = value_list

        self.obj_type = self.__class__.__name__


TemplatePropertySchema.Meta.object_class = TemplateProperty


class TemplateResourceRequirements(Object):
    """Provides the template resource requirements resource.

    Parameters
    ----------
    platform : TemplateProperty, optional
    memory : TemplateProperty, optional
    num_cores : TemplateProperty, optional
    disk_space : TemplateProperty, optional
    distributed : TemplateProperty, optional
    custom : dict[str, TemplateProperty], optional
    hpc_resources : HpcResources, optional
    """

    class Meta:
        schema = TemplateResourceRequirementsSchema
        rest_name = "None"

    def __init__(
        self,
        platform: TemplateProperty = missing,
        memory: TemplateProperty = missing,
        num_cores: TemplateProperty = missing,
        disk_space: TemplateProperty = missing,
        distributed: TemplateProperty = missing,
        custom: Dict[str, TemplateProperty] = missing,
        hpc_resources: HpcResources = missing,
        **kwargs,
    ):
        self.platform = platform
        self.memory = memory
        self.num_cores = num_cores
        self.disk_space = disk_space
        self.distributed = distributed
        self.custom = custom
        self.hpc_resources = hpc_resources

        self.obj_type = self.__class__.__name__


TemplateResourceRequirementsSchema.Meta.object_class = TemplateResourceRequirements


class TemplateInputFile(Object):
    """Provides the template input file resource.

    Parameters
    ----------
    name : str
        Name of the file.
    type : str, optional
        MIME type of the file. For example, ``text/plain``.
    evaluation_path : str, optional
        Path that the file is expected to be found under during evaluation.
    description : str
        Description of the file's purpose.
    required : bool
        Whether the file is required by the task.
    """

    class Meta:
        schema = TemplateInputFileSchema
        rest_name = "None"

    def __init__(
        self,
        name: str = missing,
        type: str = missing,
        evaluation_path: str = missing,
        description: str = missing,
        required: bool = missing,
        **kwargs,
    ):
        self.name = name
        self.type = type
        self.evaluation_path = evaluation_path
        self.description = description
        self.required = required

        self.obj_type = self.__class__.__name__


TemplateInputFileSchema.Meta.object_class = TemplateInputFile


class TemplateOutputFile(Object):
    """Provides the template output file resource.

    Parameters
    ----------
    name : str
        Name of the file.
    type : str, optional
        MIME type of the file. For example, ``text/plain``.
    evaluation_path : str, optional
        Path that the file is expected to be found under during evaluation.
    description : str
        Description of the file's purpose.
    required : bool
        Whether the file is required by the task.
    monitor : bool, optional
        Whether to live monitor the file's contents.
    collect : bool, optional
        Whether to collect files per job.
    """

    class Meta:
        schema = TemplateOutputFileSchema
        rest_name = "None"

    def __init__(
        self,
        name: str = missing,
        type: str = missing,
        evaluation_path: str = missing,
        description: str = missing,
        required: bool = missing,
        monitor: bool = missing,
        collect: bool = missing,
        **kwargs,
    ):
        self.name = name
        self.type = type
        self.evaluation_path = evaluation_path
        self.description = description
        self.required = required
        self.monitor = monitor
        self.collect = collect

        self.obj_type = self.__class__.__name__


TemplateOutputFileSchema.Meta.object_class = TemplateOutputFile


class TaskDefinitionTemplate(Object):
    """Provides the task definition template resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    modification_time : datetime, optional
        Last time in UTC (Coordinated Universal Time) that the object was modified.
    creation_time : datetime, optional
        Time in UTC when the object was created.
    name : str
        Name of the template.
    version : str, optional
        Version of the template.
    description : str, optional
        Description of the template.
    software_requirements : list[Software], optional
        List of required software.
    resource_requirements : TemplateResourceRequirements, optional
        Hardware requirements such as the number of cores, memory, and disk space.
    execution_context : dict[str, TemplateProperty], optional
        Dictionary of additional arguments to pass to the executing command.
    environment : dict[str, TemplateProperty], optional
        Dictionary of environment variables to set for the executed process.
    execution_command : str, optional
        Command to execute. A command or execution script is required.
    use_execution_script : bool, optional
        Whether to run the task with the execution script or the execution command.
    execution_script_storage_id : str, optional
        Storage ID of the script to execute (command or execution script is required).
    execution_script_storage_bucket : str, optional
        File storage bucket where the execution script is located.
    input_files : list[TemplateInputFile], optional
        List of predefined input files.
    output_files : list[TemplateOutputFile], optional
        List of predefined output files.
    """

    class Meta:
        schema = TaskDefinitionTemplateSchema
        rest_name = "task_definition_templates"

    def __init__(
        self,
        id: str = missing,
        modification_time: datetime = missing,
        creation_time: datetime = missing,
        name: str = missing,
        version: str = missing,
        description: str = missing,
        software_requirements: List[Software] = missing,
        resource_requirements: TemplateResourceRequirements = missing,
        execution_context: Dict[str, TemplateProperty] = missing,
        environment: Dict[str, TemplateProperty] = missing,
        execution_command: str = missing,
        use_execution_script: bool = missing,
        execution_script_storage_id: str = missing,
        execution_script_storage_bucket: str = missing,
        input_files: List[TemplateInputFile] = missing,
        output_files: List[TemplateOutputFile] = missing,
        **kwargs,
    ):
        self.id = id
        self.modification_time = modification_time
        self.creation_time = creation_time
        self.name = name
        self.version = version
        self.description = description
        self.software_requirements = software_requirements
        self.resource_requirements = resource_requirements
        self.execution_context = execution_context
        self.environment = environment
        self.execution_command = execution_command
        self.use_execution_script = use_execution_script
        self.execution_script_storage_id = execution_script_storage_id
        self.execution_script_storage_bucket = execution_script_storage_bucket
        self.input_files = input_files
        self.output_files = output_files

        self.obj_type = self.__class__.__name__


TaskDefinitionTemplateSchema.Meta.object_class = TaskDefinitionTemplate



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing restricted value fields."""
from marshmallow import fields
from marshmallow.exceptions import ValidationError


class RestrictedValue(fields.Field):
    """Restricted value fields."""

    restricted_fields = [
        fields.Int(strict=True),
        fields.Bool(truthy=[True], falsy=[False]),
        fields.Str(),
        fields.Float(allow_nan=False),
    ]

    def __init__(self):
        super().__init__(allow_none=True)

    def _deserialize(self, value, attr, obj, **kwargs):
        """Convert string to restricted value object."""
        for field in self.restricted_fields:
            try:
                return field._deserialize(value, attr, obj, **kwargs)
            except:
                pass

        self.raise_validation_error()

    def raise_validation_error():
        """Raise validation error if value is not a float, integer, Boolean, or string."""
        raise ValidationError("Value must be a float, integer, Boolean, or string.")



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing template property, resource requirements, file, input file,
output file, and task definition schema."""
from marshmallow import fields, validate

from ansys.hps.client.common import BaseSchema, ObjectSchema

from .task_definition import HpcResourcesSchema, SoftwareSchema


class TemplatePropertySchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    default = fields.Raw(allow_none=True, metadata={"description": "Default value."})
    description = fields.String(
        allow_none=True, metadata={"description": "Description of the property's purpose."}
    )
    type = fields.String(
        allow_none=True,
        validate=validate.OneOf(["int", "float", "bool", "string"]),
        metadata={
            "description": "Type of the property. "
            "Options are ``bool``, ``float``, ``int``, and ``string``."
        },
    )
    value_list = fields.Raw(
        allow_none=True,
        dump_default=[],
        load_default=[],
        metadata={"many": True, "description": "List of possible values for the property."},
    )


class TemplateResourceRequirementsSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    platform = fields.Nested(TemplatePropertySchema, allow_none=True)
    memory = fields.Nested(TemplatePropertySchema, allow_none=True)
    num_cores = fields.Nested(TemplatePropertySchema, allow_none=True)
    disk_space = fields.Nested(TemplatePropertySchema, allow_none=True)
    distributed = fields.Nested(TemplatePropertySchema, allow_none=True)
    custom = fields.Dict(
        keys=fields.String, values=fields.Nested(TemplatePropertySchema), allow_none=True
    )
    hpc_resources = fields.Nested(HpcResourcesSchema, allow_none=True)


class TemplateFileSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        pass

    name = fields.String(metadata={"description": "Name of the file."})
    type = fields.String(
        allow_none=True,
        metadata={"description": "MIME type of the file. For example, ``text/plain``."},
    )
    evaluation_path = fields.String(
        allow_none=True,
        metadata={
            "description": "Path that the file is expected to be found under during evaluation."
        },
    )
    description = fields.String(metadata={"description": "Description of the file's purpose."})
    required = fields.Bool(metadata={"description": "Whether the file is required by the task."})


class TemplateInputFileSchema(TemplateFileSchema):
    pass


class TemplateOutputFileSchema(TemplateFileSchema):
    monitor = fields.Bool(
        allow_none=True, metadata={"description": "Whether to live monitor the file's contents."}
    )
    collect = fields.Bool(
        allow_none=True, metadata={"description": "Whether to collect files per job."}
    )


class TaskDefinitionTemplateSchema(ObjectSchema):
    class Meta(ObjectSchema.Meta):
        pass

    modification_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={
            "description": "Last time in UTC (Coordinated Universal Time) "
            "that the object was modified."
        },
    )
    creation_time = fields.DateTime(
        allow_none=True,
        load_only=True,
        metadata={"description": "Time in UTC when the object was created."},
    )

    name = fields.String(metadata={"description": "Name of the template."})
    version = fields.String(metadata={"description": "Version of the template."}, allow_none=True)
    description = fields.String(
        metadata={"description": "Description of the template."}, allow_none=True
    )

    software_requirements = fields.Nested(
        SoftwareSchema,
        many=True,
        allow_none=True,
        metadata={"description": "List of required software."},
    )
    resource_requirements = fields.Nested(
        TemplateResourceRequirementsSchema,
        allow_none=True,
        metadata={
            "description": "Hardware requirements such as the number of cores, "
            "memory, and disk space."
        },
    )

    execution_context = fields.Dict(
        keys=fields.String,
        values=fields.Nested(TemplatePropertySchema),
        allow_none=True,
        metadata={
            "description": "Dictionary of additional arguments to pass to the executing command."
        },
    )
    environment = fields.Dict(
        keys=fields.String,
        values=fields.Nested(TemplatePropertySchema),
        allow_none=True,
        metadata={
            "description": "Dictionary of environment variables to set for the executed process."
        },
    )

    execution_command = fields.String(
        allow_none=True,
        metadata={"description": "Command to execute. A command or execution script is required."},
    )
    use_execution_script = fields.Bool(
        allow_none=True,
        metadata={
            "description": "Whether to run the task with the execution script "
            "or the execution command."
        },
    )
    execution_script_storage_id = fields.String(
        allow_none=True,
        metadata={
            "description": "Storage ID of the script to execute "
            "(command or execution script is required).",
        },
    )
    execution_script_storage_bucket = fields.String(
        allow_none=True,
        metadata={"description": "File storage bucket where the execution script is located."},
    )

    input_files = fields.Nested(
        TemplateInputFileSchema,
        many=True,
        allow_none=True,
        metadata={"description": "List of predefined input files."},
    )
    output_files = fields.Nested(
        TemplateOutputFileSchema,
        many=True,
        allow_none=True,
        metadata={"description": "List of predefined output files."},
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing job schema."""

from marshmallow import fields
from marshmallow.validate import OneOf

from ansys.hps.client.common import ObjectSchemaWithModificationInfo

from .object_reference import IdReference, IdReferenceList

valid_eval_status = [
    "inactive",
    "pending",
    "prolog",
    "running",
    "evaluated",
    "failed",
    "aborted",
    "timeout",
]


class JobSchema(ObjectSchemaWithModificationInfo):
    class Meta(ObjectSchemaWithModificationInfo.Meta):
        pass

    name = fields.String(allow_none=True, metadata={"description": "Name of the job."})
    eval_status = fields.String(
        validate=OneOf(valid_eval_status), metadata={"description": "Evaluation status."}
    )
    job_definition_id = IdReference(
        allow_none=False,
        attribute="job_definition_id",
        referenced_class="JobDefinition",
        metadata={
            "description": "ID of the linked job definition. "
            "For more information, see the :class:`JobDefinition` class."
        },
    )

    priority = fields.Integer(
        allow_none=True,
        metadata={
            "description": "Priority for evaluating the job. "
            "The default is ``0``, which is the highest priority. "
            "Assigning a higher value to a job makes it a lower priority."
        },
    )
    values = fields.Dict(
        keys=fields.String(),
        allow_none=True,
        metadata={
            "description": "Dictionary with (name,value) pairs for all parameters defined in the "
            "linked job definition."
        },
    )
    fitness = fields.Float(allow_none=True, metadata={"description": "Fitness value computed."})
    fitness_term_values = fields.Dict(
        keys=fields.String(),
        values=fields.Float(allow_none=True),
        allow_none=True,
        metadata={
            "description": "Dictionary with (name,value) pairs for all fitness terms computed."
        },
    )
    note = fields.String(allow_none=True, metadata={"description": "Note for the job."})
    creator = fields.String(
        allow_none=True,
        metadata={"description": "Additional information about the creator of the job."},
    )

    executed_level = fields.Integer(
        allow_none=True,
        metadata={
            "description": "Execution level of the last executed task. "
            "A value of ``-1`` indicates that no task has been executed yet."
        },
    )

    elapsed_time = fields.Float(
        load_only=True,
        metadata={"description": "Number of seconds it took the evaluators to update the job."},
    )

    host_ids = fields.List(
        fields.String(allow_none=True),
        allow_none=True,
        metadata={"description": "List of host IDs of the evaluators that updated the job."},
    )
    file_ids = IdReferenceList(
        referenced_class="File",
        attribute="file_ids",
        load_only=True,
        metadata={"description": "List of IDs of all files of the job."},
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the task definition resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.task_definition import (
    HpcResourcesSchema,
    LicensingSchema,
    ResourceRequirementsSchema,
    SoftwareSchema,
    SuccessCriteriaSchema,
    TaskDefinitionSchema,
)


class HpcResources(Object):
    """Provides the hpc resources resource.

    Parameters
    ----------
    num_cores_per_node : int, optional
        Number of cores per node.
    num_gpus_per_node : int, optional
        Number of GPUs per node.
    exclusive : bool, optional
        Whether a job can't share resources with other running jobs.
    queue : str, optional
        Name of the job scheduler queue.
    """

    class Meta:
        schema = HpcResourcesSchema
        rest_name = "None"

    def __init__(
        self,
        num_cores_per_node: int = missing,
        num_gpus_per_node: int = missing,
        exclusive: bool = missing,
        queue: str = missing,
        **kwargs,
    ):
        self.num_cores_per_node = num_cores_per_node
        self.num_gpus_per_node = num_gpus_per_node
        self.exclusive = exclusive
        self.queue = queue

        self.obj_type = self.__class__.__name__


HpcResourcesSchema.Meta.object_class = HpcResources


class ResourceRequirements(Object):
    """Provides the resource requirements resource.

    Parameters
    ----------
    platform : str, optional
        Basic platform information. Options are ``'linux'`` and ``'windows'``.
    memory : int, optional
        Amount of RAM in bytes.
    num_cores : float, optional
        Number of cores.
    disk_space : int, optional
        Amount of disk space in bytes.
    distributed : bool, optional
        Whether to enable distributed parallel processing.
    custom : dict[str, Union[int, float, str, bool]], optional
        Dictionary of custom resource requirements.
    hpc_resources : HpcResources, optional
        HPC resource requirements.
    """

    class Meta:
        schema = ResourceRequirementsSchema
        rest_name = "None"

    def __init__(
        self,
        platform: str = missing,
        memory: int = missing,
        num_cores: float = missing,
        disk_space: int = missing,
        distributed: bool = missing,
        custom: Dict[str, Union[int, float, str, bool]] = missing,
        hpc_resources: HpcResources = missing,
        **kwargs,
    ):
        self.platform = platform
        self.memory = memory
        self.num_cores = num_cores
        self.disk_space = disk_space
        self.distributed = distributed
        self.custom = custom
        self.hpc_resources = hpc_resources

        self.obj_type = self.__class__.__name__


ResourceRequirementsSchema.Meta.object_class = ResourceRequirements


class Software(Object):
    """Provides the software resource.

    Parameters
    ----------
    name : str
        Name of the app.
    version : str, optional
        Version of the app.
    """

    class Meta:
        schema = SoftwareSchema
        rest_name = "None"

    def __init__(self, name: str = missing, version: str = missing, **kwargs):
        self.name = name
        self.version = version

        self.obj_type = self.__class__.__name__


SoftwareSchema.Meta.object_class = Software


class SuccessCriteria(Object):
    """Provides the success criteria resource.

    Parameters
    ----------
    return_code : int, optional
        Process exit code that must be returned by the executed command.
    expressions : list, optional
        List of expressions to evaluate.
    required_output_file_ids : list[str], optional
        List of IDs of the required output files.
    require_all_output_files : bool, optional
        Whether to require all output files.
    required_output_parameter_ids : list[str], optional
        List of names of the required output parameters.
    require_all_output_parameters : bool, optional
        Whether to require all output parameters.
    """

    class Meta:
        schema = SuccessCriteriaSchema
        rest_name = "None"

    def __init__(
        self,
        return_code: int = missing,
        expressions: List = missing,
        required_output_file_ids: List[str] = missing,
        require_all_output_files: bool = missing,
        required_output_parameter_ids: List[str] = missing,
        require_all_output_parameters: bool = missing,
        **kwargs,
    ):
        self.return_code = return_code
        self.expressions = expressions
        self.required_output_file_ids = required_output_file_ids
        self.require_all_output_files = require_all_output_files
        self.required_output_parameter_ids = required_output_parameter_ids
        self.require_all_output_parameters = require_all_output_parameters

        self.obj_type = self.__class__.__name__


SuccessCriteriaSchema.Meta.object_class = SuccessCriteria


class Licensing(Object):
    """Provides the licensing resource.

    Parameters
    ----------
    enable_shared_licensing : bool, optional
        Whether to enable shared licensing contexts for Ansys simulations.
    """

    class Meta:
        schema = LicensingSchema
        rest_name = "None"

    def __init__(self, enable_shared_licensing: bool = missing, **kwargs):
        self.enable_shared_licensing = enable_shared_licensing

        self.obj_type = self.__class__.__name__


LicensingSchema.Meta.object_class = Licensing


class TaskDefinition(Object):
    """Provides the task definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name.
    execution_command : str, optional
        Command to execute. A command or execution script is required.
    use_execution_script : bool, optional
        Whether to run the task with the execution script or the execution command.
    execution_script_id : str, optional
        Script to execute. A command or execution script is required.
    execution_level : int
        Execution level for the task.
    execution_context : dict[str, Union[int, float, str, bool]], optional
        Additional arguments to pass to the executing command.
    environment : dict[str, str], optional
        Environment variables to set for the executed process.
    max_execution_time : float, optional
        Maximum time in seconds for executing the task.
    num_trials : int, optional
        Maximum number of attempts for executing the task.
    store_output : bool, optional
        Whether to store the standard output of the task.
    input_file_ids : list[str]
        List of IDs of input files.
    output_file_ids : list[str]
        List of IDs of output files.
    success_criteria : SuccessCriteria, optional
    licensing : Licensing, optional
        :class:`Licensing` object.
    software_requirements : list[Software], optional
        List of :class:`Software` objects.
    resource_requirements : ResourceRequirements, optional
        :class:`ResourceRequirements` object.
    """

    class Meta:
        schema = TaskDefinitionSchema
        rest_name = "task_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        execution_command: str = missing,
        use_execution_script: bool = missing,
        execution_script_id: str = missing,
        execution_level: int = missing,
        execution_context: Dict[str, Union[int, float, str, bool]] = missing,
        environment: Dict[str, str] = missing,
        max_execution_time: float = missing,
        num_trials: int = missing,
        store_output: bool = missing,
        input_file_ids: List[str] = missing,
        output_file_ids: List[str] = missing,
        success_criteria: SuccessCriteria = missing,
        licensing: Licensing = missing,
        software_requirements: List[Software] = missing,
        resource_requirements: ResourceRequirements = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.execution_command = execution_command
        self.use_execution_script = use_execution_script
        self.execution_script_id = execution_script_id
        self.execution_level = execution_level
        self.execution_context = execution_context
        self.environment = environment
        self.max_execution_time = max_execution_time
        self.num_trials = num_trials
        self.store_output = store_output
        self.input_file_ids = input_file_ids
        self.output_file_ids = output_file_ids
        self.success_criteria = success_criteria
        self.licensing = licensing
        self.software_requirements = software_requirements
        self.resource_requirements = resource_requirements

        self.obj_type = self.__class__.__name__


TaskDefinitionSchema.Meta.object_class = TaskDefinition



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the parameter definition resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.parameter_definition import (
    BoolParameterDefinitionSchema,
    FloatParameterDefinitionSchema,
    IntParameterDefinitionSchema,
    ParameterDefinitionSchema,
    StringParameterDefinitionSchema,
)


class ParameterDefinition(Object):
    """Provides the parameter definition resource.

    Parameters
    ----------

    """

    class Meta:
        schema = ParameterDefinitionSchema
        rest_name = "parameter_definitions"

    def __init__(self, **kwargs):

        self.obj_type = self.__class__.__name__


ParameterDefinitionSchema.Meta.object_class = ParameterDefinition


class FloatParameterDefinition(ParameterDefinition):
    """Provides the float parameter definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name (ID) of the parameter.
    quantity_name : str, optional
        Name of the quantity that the parameter represents. For example, ``Length``.
    units : str, optional
        Units for the parameter.
    display_text : str, optional
        Text to display as the parameter name.
    mode : str
        Indicates whether it's an input or output parameter. The mode is filled server-side.
    type : str
    default : float, optional
        Default parameter value.
    lower_limit : float, optional
        Lower bound for the parameter value.
    upper_limit : float, optional
        Upper bound for the parameter value.
    step : float, optional
        If provided, allowable values are given by: AllowableValue = lower_limit + n * step, where n is an integer and AllowableValue <= upper_limit.
    cyclic : bool, optional
        Whether the parameter is cyclic.
    value_list : list, optional
        List of allowed values. This parameter provides an alternative to specifying upper and lower limits.
    """

    class Meta:
        schema = FloatParameterDefinitionSchema
        rest_name = "parameter_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        quantity_name: str = missing,
        units: str = missing,
        display_text: str = missing,
        mode: str = missing,
        type: str = missing,
        default: float = missing,
        lower_limit: float = missing,
        upper_limit: float = missing,
        step: float = missing,
        cyclic: bool = missing,
        value_list: List = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.quantity_name = quantity_name
        self.units = units
        self.display_text = display_text
        self.mode = mode
        self.type = type
        self.default = default
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.step = step
        self.cyclic = cyclic
        self.value_list = value_list

        self.obj_type = self.__class__.__name__


FloatParameterDefinitionSchema.Meta.object_class = FloatParameterDefinition


class IntParameterDefinition(ParameterDefinition):
    """Provides the int parameter definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name (ID) of the parameter.
    quantity_name : str, optional
        Name of the quantity that the parameter represents. For example, ``Length``.
    units : str, optional
        Units for the parameter.
    display_text : str, optional
        Text to display as the parameter name.
    mode : str
        Indicates whether it's an input or output parameter. The mode is filled server-side.
    type : str
    default : int, optional
        Default parameter value.
    lower_limit : int, optional
        Lower bound for the parameter value.
    upper_limit : int, optional
        Upper bound for the parameter value.
    step : int, optional
        The default is ``1``.
    cyclic : bool, optional
        Whether the parameter is cyclic.
    """

    class Meta:
        schema = IntParameterDefinitionSchema
        rest_name = "parameter_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        quantity_name: str = missing,
        units: str = missing,
        display_text: str = missing,
        mode: str = missing,
        type: str = missing,
        default: int = missing,
        lower_limit: int = missing,
        upper_limit: int = missing,
        step: int = missing,
        cyclic: bool = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.quantity_name = quantity_name
        self.units = units
        self.display_text = display_text
        self.mode = mode
        self.type = type
        self.default = default
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.step = step
        self.cyclic = cyclic

        self.obj_type = self.__class__.__name__


IntParameterDefinitionSchema.Meta.object_class = IntParameterDefinition


class BoolParameterDefinition(ParameterDefinition):
    """Provides the bool parameter definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name (ID) of the parameter.
    quantity_name : str, optional
        Name of the quantity that the parameter represents. For example, ``Length``.
    units : str, optional
        Units for the parameter.
    display_text : str, optional
        Text to display as the parameter name.
    mode : str
        Indicates whether it's an input or output parameter. The mode is filled server-side.
    type : str
    default : bool, optional
        Default parameter value.
    """

    class Meta:
        schema = BoolParameterDefinitionSchema
        rest_name = "parameter_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        quantity_name: str = missing,
        units: str = missing,
        display_text: str = missing,
        mode: str = missing,
        type: str = missing,
        default: bool = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.quantity_name = quantity_name
        self.units = units
        self.display_text = display_text
        self.mode = mode
        self.type = type
        self.default = default

        self.obj_type = self.__class__.__name__


BoolParameterDefinitionSchema.Meta.object_class = BoolParameterDefinition


class StringParameterDefinition(ParameterDefinition):
    """Provides the string parameter definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name (ID) of the parameter.
    quantity_name : str, optional
        Name of the quantity that the parameter represents. For example, ``Length``.
    units : str, optional
        Units for the parameter.
    display_text : str, optional
        Text to display as the parameter name.
    mode : str
        Indicates whether it's an input or output parameter. The mode is filled server-side.
    type : str
    default : str, optional
        Default parameter value.
    value_list : list, optional
        List of allowed values.
    """

    class Meta:
        schema = StringParameterDefinitionSchema
        rest_name = "parameter_definitions"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        quantity_name: str = missing,
        units: str = missing,
        display_text: str = missing,
        mode: str = missing,
        type: str = missing,
        default: str = missing,
        value_list: List = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.quantity_name = quantity_name
        self.units = units
        self.display_text = display_text
        self.mode = mode
        self.type = type
        self.default = default
        self.value_list = value_list

        self.obj_type = self.__class__.__name__


StringParameterDefinitionSchema.Meta.object_class = StringParameterDefinition



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the user resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.user import UserSchema


class User(Object):
    """Provides the user resource.

    Parameters
    ----------
    id : str
        Unique user ID, generated internally by the server on creation.
    username : str
        Username.
    password : str
        Password.
    first_name : str, optional
        First name.
    last_name : str, optional
        Last name.
    email : str, optional
        E-mail address.
    """

    class Meta:
        schema = UserSchema
        rest_name = "None"

    def __init__(
        self,
        id: str = missing,
        username: str = missing,
        password: str = missing,
        first_name: str = missing,
        last_name: str = missing,
        email: str = missing,
        **kwargs,
    ):
        self.id = id
        self.username = username
        self.password = password
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

        self.obj_type = self.__class__.__name__


UserSchema.Meta.object_class = User



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the parameter mapping resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.parameter_mapping import ParameterMappingSchema


class ParameterMapping(Object):
    """Provides the parameter mapping resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    line : int, optional
    column : int, optional
    key_string : str, optional
    float_field : str, optional
    width : int, optional
    precision : int, optional
    tokenizer : str, optional
    decimal_symbol : str, optional
    digit_grouping_symbol : str, optional
    string_quote : str, optional
    true_string : str, optional
    false_string : str, optional
    parameter_definition_id : str, optional
        ID of the linked parameter definition. For more information, see the :class:`ParameterDefinition` class.
    task_definition_property : str, optional
    file_id : str, optional
        ID of the file resource.
    """

    class Meta:
        schema = ParameterMappingSchema
        rest_name = "parameter_mappings"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        line: int = missing,
        column: int = missing,
        key_string: str = missing,
        float_field: str = missing,
        width: int = missing,
        precision: int = missing,
        tokenizer: str = missing,
        decimal_symbol: str = missing,
        digit_grouping_symbol: str = missing,
        string_quote: str = missing,
        true_string: str = missing,
        false_string: str = missing,
        parameter_definition_id: str = missing,
        task_definition_property: str = missing,
        file_id: str = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.line = line
        self.column = column
        self.key_string = key_string
        self.float_field = float_field
        self.width = width
        self.precision = precision
        self.tokenizer = tokenizer
        self.decimal_symbol = decimal_symbol
        self.digit_grouping_symbol = digit_grouping_symbol
        self.string_quote = string_quote
        self.true_string = true_string
        self.false_string = false_string
        self.parameter_definition_id = parameter_definition_id
        self.task_definition_property = task_definition_property
        self.file_id = file_id

        self.obj_type = self.__class__.__name__


ParameterMappingSchema.Meta.object_class = ParameterMapping



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the license context resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.license_context import LicenseContextSchema


class LicenseContext(Object):
    """Provides the license context resource.

    Parameters
    ----------
    context_id : str, optional
        License context ID.
    environment : dict, optional
        License context environment dict.
    """

    class Meta:
        schema = LicenseContextSchema
        rest_name = "license_contexts"

    def __init__(self, context_id: str = missing, environment: Dict = missing, **kwargs):
        self.context_id = context_id
        self.environment = environment

        self.obj_type = self.__class__.__name__


LicenseContextSchema.Meta.object_class = LicenseContext



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the algorithm resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.algorithm import AlgorithmSchema


class Algorithm(Object):
    """Provides the algorithm resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    creation_time : datetime, optional
        Date and time that the resource was created.
    modification_time : datetime, optional
        Date and time that the resource was last modified.
    created_by : str, optional
        ID of the user who created the object.
    modified_by : str, optional
        ID of the user who last modified the object.
    name : str, optional
        Name of the algorithm.
    description : str, optional
        Description of the algorithm.
    data : str, optional
        Generic string field to hold arbitrary algorithm configuration data. For example, as a JSON dictionary.
    jobs : list[str]
        List of job IDs.
    """

    class Meta:
        schema = AlgorithmSchema
        rest_name = "algorithms"

    def __init__(
        self,
        id: str = missing,
        creation_time: datetime = missing,
        modification_time: datetime = missing,
        created_by: str = missing,
        modified_by: str = missing,
        name: str = missing,
        description: str = missing,
        data: str = missing,
        jobs: List[str] = missing,
        **kwargs,
    ):
        self.id = id
        self.creation_time = creation_time
        self.modification_time = modification_time
        self.created_by = created_by
        self.modified_by = modified_by
        self.name = name
        self.description = description
        self.data = data
        self.jobs = jobs

        self.obj_type = self.__class__.__name__


AlgorithmSchema.Meta.object_class = Algorithm



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing the fitness definition and fitness term definition resources."""
import logging
from typing import List

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.fitness_definition import FitnessDefinitionSchema, FitnessTermDefinitionSchema

log = logging.getLogger(__name__)


class FitnessTermDefinition(Object):
    """Provides the fitness term definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    name : str, optional
        Name of the fitness term.
    expression : str, optional
        Python expression that defines the fitness term.
    type : str, optional
        Fitness term type.
    weighting_factor : float, optional
        Relative importance of the fitness term in comparison to other fitness terms.

    Examples
    --------

    Create a fitness term of the objective type.

    >>> ft1 = FitnessTermDefinition(
    ...     name="weight",
    ...     type="design_objective",
    ...     weighting_factor=1.0,
    ...     expression="map_design_objective(values['weight'],7.5,5.5)"
    ... )

    Create a fitness term of the target constraint type.

    >>> ft2 = FitnessTermDefinition(
    ...     name="torsional_stiffness",
    ...     type="target_constraint",
    ...     weighting_factor=0.8,
    ...     expression="map_target_constraint(
    ...         values['torsion_stiffness'], 1313.0, 5.0, 30.0)"
    ... )

    Create a fitness term of the limit constraint type.

    >>> ft3 = FitnessTermDefinition(
    ...     name="max_stress",
    ...     type="limit_constraint",
    ...     weighting_factor=0.6,
    ...     expression="map_limit_constraint(
    ...         values['max_stress'], 451.0, 50.0 )"
    ... )
    """

    class Meta:
        schema = FitnessTermDefinitionSchema
        rest_name = "None"

    def __init__(
        self,
        id: str = missing,
        name: str = missing,
        expression: str = missing,
        type: str = missing,
        weighting_factor: float = missing,
        **kwargs,
    ):
        self.id = id
        self.name = name
        self.expression = expression
        self.type = type
        self.weighting_factor = weighting_factor

        self.obj_type = self.__class__.__name__


FitnessTermDefinitionSchema.Meta.object_class = FitnessTermDefinition


class FitnessDefinition(Object):
    """Provides the fitness definition resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    fitness_term_definitions
        List of :class:`ansys.hps.client.jms.FitnessTermDefinition` objects.
    error_fitness : float
        Default fitness value to assign to failed jobs.

    Examples
    --------
    >>> fd = FitnessDefinition(error_fitness=10.0)
    >>> fd.add_fitness_term(name="weight", type="design_objective", weighting_factor=1.0,
    ...         expression="map_design_objective( values['weight'], 7.5, 5.5)")
    >>> fd.add_fitness_term(name="torsional_stiffness",
    ...                     type="target_constraint",
    ...                     weighting_factor=1.0,
    ...                     expression="map_target_constraint(
    ...                         values['torsion_stiffness'],
    ...                         1313.0,
    ...                         5.0,
    ...                         30.0 )"
    ...                     )
    """

    class Meta:
        schema = FitnessDefinitionSchema
        rest_name = "None"

    def __init__(
        self,
        id: str = missing,
        fitness_term_definitions: List[FitnessTermDefinition] = missing,
        error_fitness: float = missing,
        **kwargs,
    ):
        self.id = id
        self.fitness_term_definitions = fitness_term_definitions
        self.error_fitness = error_fitness

        self.obj_type = self.__class__.__name__

    def add_fitness_term(self, **kwargs):
        """
        Add a fitness term easily using a helper function.
        """
        ft = FitnessTermDefinition(**kwargs)

        if self.fitness_term_definitions == missing:
            self.fitness_term_definitions = []
        self.fitness_term_definitions.append(ft)
        return ft


FitnessDefinitionSchema.Meta.object_class = FitnessDefinition



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing the Python client to the HPS APIs."""

import logging
from typing import Union
import warnings

import jwt
import requests

from .authenticate import authenticate
from .connection import create_session
from .exceptions import HPSError, raise_for_status
from .warnings import UnverifiedHTTPSRequestsWarning

log = logging.getLogger(__name__)


class Client(object):
    """Provides the Python client to the HPS APIs.

    This class uses the provided credentials to create and store
    an authorized :class:`requests.Session` object.

    The following alternative authentication workflows are supported
    and evaluated in the order listed:

    - Access token: No authentication is needed.
    - Username and password: The client connects to the OAuth server and
      requests access and refresh tokens.
    - Refresh token: The client connects to the OAuth server and
      requests a new access token.
    - Client credentials: The client authenticates with the ``client_id`` and ``client_secret``
      parameters to obtain a new access token. (A refresh token is not included.)

    Parameters
    ----------
    url : str
        Base path for the server to call. The default is ``'https://127.0.0.1:8443/hps'``.
    username : str, optional
        Username.
    password : str, optional
        Password.
    realm : str, optional
        Keycloak realm. The default is ``'rep'``.
    grant_type : str, optional
        Authentication method. The default is ``'password'``.
    scope : str, optional
        String containing one or more requested scopes. The default is ``'openid'``.
    client_id : str, optional
        Client type. The default is ``'rep-cli'``.
    client_secret : str, optional
        Client secret. The default is ``None``.
    access_token : str, optional
        Access token.
    refresh_token : str, optional
        Refresh token.
    auth_url : str, optional
    all_fields : bool, optional
        Whether to apply the ``fields="all"`` query parameter to all requests so
        that all available fields are returned for the requested resources. The
        default is ``True``.
    verify : Union[bool, str], optional
        If a Boolean, whether to verify the server's TLS certificate. The default
        is ``None`, which disables certificate validation and warns the user about it.
        If a string, the path to the CA bundle to use. For more information,
        see the :class:`requests.Session` documentation.
    disable_security_warnings : bool, optional
        Whether to disable urllib3 warnings about insecure HTTPS requests. The default is ``True``.
        For more information, see urllib3 documentation about TLS warnings.

    Examples
    --------

    Create a client object and connect to HPS with a username and password.

    >>> from ansys.hps.client import Client
    >>> cl = Client(
    ...     url="https://localhost:8443/hps",
    ...     username="repuser",
    ...     password="repuser"
    ... )

    Create a client object and connect to HPS with a refresh token.

    >>> cl = Client(
    ...     url="https://localhost:8443/hps",
    ...     username="repuser",
    ...     refresh_token="eyJhbGciOiJIUzI1NiIsInR5cC..."
    >>> )

    """

    def __init__(
        self,
        url: str = "https://127.0.0.1:8443/hps",
        username: str = None,
        password: str = None,
        *,
        realm: str = "rep",
        grant_type: str = None,
        scope="openid",
        client_id: str = "rep-cli",
        client_secret: str = None,
        access_token: str = None,
        refresh_token: str = None,
        auth_url: str = None,
        all_fields=True,
        verify: Union[bool, str] = None,
        disable_security_warnings: bool = True,
        **kwargs,
    ):

        rep_url = kwargs.get("rep_url", None)
        if rep_url is not None:
            url = rep_url
            msg = "The 'rep_url' input argument is deprecated. Use 'url' instead."
            warnings.warn(msg, DeprecationWarning)
            log.warning(msg)

        self.url = url
        self.auth_url = auth_url
        self.auth_api_url = (auth_url or url) + f"/auth/"
        self.access_token = None
        self.refresh_token = None
        self.username = username
        self.realm = realm
        self.grant_type = grant_type
        self.scope = scope
        self.client_id = client_id
        self.client_secret = client_secret
        self.verify = verify

        if self.verify is None:
            self.verify = False
            msg = (
                f"Certificate verification is disabled. "
                f"Unverified HTTPS requests are made to {self.url}."
            )
            warnings.warn(msg, UnverifiedHTTPSRequestsWarning)
            log.warning(msg)

        if disable_security_warnings:
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning
            )

        if access_token:
            log.debug("Authenticate with access token")
            self.access_token = access_token
        else:
            if username and password:
                self.grant_type = "password"
            elif refresh_token:
                self.grant_type = "refresh_token"
            elif client_secret:
                self.grant_type = "client_credentials"

            log.debug(f"Authenticating with '{self.grant_type}' grant type.")

            tokens = authenticate(
                url=auth_url or url,
                realm=realm,
                grant_type=self.grant_type,
                scope=scope,
                client_id=client_id,
                client_secret=client_secret,
                username=username,
                password=password,
                refresh_token=refresh_token,
                verify=self.verify,
            )
            self.access_token = tokens["access_token"]
            # client credentials flow does not return a refresh token
            self.refresh_token = tokens.get("refresh_token", None)

        parsed_username = None

        try:
            parsed_jwt = jwt.decode(self.access_token, options={"verify_signature": False})
            parsed_username = parsed_jwt["preferred_username"]
        except:
            log.warning("Could not retrieve preferred_username from access token.")

        if parsed_username is not None:
            if self.username is not None and self.username != parsed_username:
                raise HPSError(
                    (
                        f"Username: '{self.username}' and "
                        f"preferred_username: '{parsed_username}' "
                        "from access token do not match."
                    )
                )
            self.username = parsed_username

        self.session = create_session(
            self.access_token,
            verify=self.verify,
        )
        if all_fields:
            self.session.params = {"fields": "all"}

        # register hook to handle expiring of the refresh token
        self.session.hooks["response"] = [self._auto_refresh_token, raise_for_status]
        self._unauthorized_num_retry = 0
        self._unauthorized_max_retry = 1

    @property
    def rep_url(self) -> str:
        msg = "The client 'rep_url' property is deprecated. Use 'url' instead."
        warnings.warn(msg, DeprecationWarning)
        log.warning(msg)
        return self.url

    def _auto_refresh_token(self, response, *args, **kwargs):
        """Automatically refreshes the access token and
        resends the request in case of an unauthorized error."""
        if (
            response.status_code == 401
            and self._unauthorized_num_retry < self._unauthorized_max_retry
        ):
            log.info(f"401 authorization error: Trying to get a new access token.")
            self._unauthorized_num_retry += 1
            self.refresh_access_token()
            response.request.headers.update(
                {"Authorization": self.session.headers["Authorization"]}
            )
            log.debug(f"Retrying request with updated access token.")
            return self.session.send(response.request)

        self._unauthorized_num_retry = 0
        return response

    def refresh_access_token(self):
        """Request a new access token"""
        if self.grant_type == "client_credentials":
            # Its not recommended to give refresh tokens to client_credentials grant types
            # as per OAuth 2.0 RFC6749 Section 4.4.3, so handle these specially...
            tokens = authenticate(
                url=self.auth_url or self.url,
                realm=self.realm,
                grant_type="client_credentials",
                scope=self.scope,
                client_id=self.client_id,
                client_secret=self.client_secret,
                verify=self.verify,
            )
        else:
            # Other workflows for authentication generally support refresh_tokens
            tokens = authenticate(
                url=self.auth_url or self.url,
                realm=self.realm,
                grant_type="refresh_token",
                scope=self.scope,
                client_id=self.client_id,
                client_secret=self.client_secret,
                username=self.username,
                refresh_token=self.refresh_token,
                verify=self.verify,
            )
        self.access_token = tokens["access_token"]
        self.refresh_token = tokens.get("refresh_token", None)
        self.session.headers.update({"Authorization": "Bearer %s" % tokens["access_token"]})



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module getting, updating, creating, and deleting objects, and converting objects
to and from JSONs."""
import json
import logging
from typing import List, Type

from pydantic import BaseModel
from pydantic import __version__ as pydantic_version
from pydantic import create_model
from requests import Session

from ansys.hps.client.exceptions import ClientError

from ..models import (
    ComputeResourceSet,
    EvaluatorConfigurationUpdate,
    EvaluatorRegistration,
    ScalerRegistration,
)

OBJECT_TYPE_TO_ENDPOINT = {
    EvaluatorRegistration: "evaluators",
    EvaluatorConfigurationUpdate: "configuration_updates",
    ScalerRegistration: "scalers",
    ComputeResourceSet: "compute_resource_sets",
}

log = logging.getLogger(__name__)


def _create_dynamic_list_model(name, field_name, field_type) -> BaseModel:
    # Helper function to create at runtime a pydantic model storing
    # a list of objects.
    fields = {f"{field_name}": (List[field_type], ...)}
    return create_model(name, **fields)


def objects_to_json(
    objects: List[BaseModel],
    rest_name: str,
    exclude_unset: bool = True,
    exclude_defaults: bool = False,
):

    ListOfObjects = _create_dynamic_list_model(
        name=f"List{objects[0].__class__.__name__}",
        field_name=rest_name,
        field_type=objects[0].__class__,
    )

    args = {f"{rest_name}": objects}
    objects_list = ListOfObjects(**args)

    if pydantic_version.startswith("1."):
        return objects_list.json(exclude_unset=exclude_unset, exclude_defaults=exclude_defaults)
    elif pydantic_version.startswith("2."):
        return objects_list.model_dump_json(
            exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
        )
    else:
        raise RuntimeError(f"Unsupported Pydantic version {pydantic_version}")


def json_to_objects(data, obj_type):
    obj_list = []
    for obj in data:
        obj_list.append(obj_type(**obj))
    return obj_list


def get_objects(
    session: Session, url: str, obj_type: Type[BaseModel], as_objects=True, **query_params
):

    rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]
    url = f"{url}/{rest_name}"
    r = session.get(url, params=query_params)

    data = r.json()[rest_name]
    if not as_objects:
        return data

    return json_to_objects(data, obj_type)


def get_objects_count(session: Session, url: str, obj_type: Type[BaseModel], **query_params):

    rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]
    url = f"{url}/{rest_name}:count"  # noqa: E231
    r = session.get(url, params=query_params)

    return r.json()[f"num_{rest_name}"]


def get_object(
    session: Session,
    url: str,
    obj_type: Type[BaseModel],
    as_object=True,
    from_collection=False,
    **query_params,
):

    r = session.get(url, params=query_params)
    data = r.json()
    if from_collection:
        rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]
        data = data[rest_name][0]
    if not as_object:
        return data
    return obj_type(**data)


def create_objects(
    session: Session, url: str, objects: List[BaseModel], as_objects=True, **query_params
):
    if not objects:
        return []

    are_same = [o.__class__ == objects[0].__class__ for o in objects[1:]]
    if not all(are_same):
        raise ClientError("Mixed object types")

    obj_type = objects[0].__class__
    rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]

    url = f"{url}/{rest_name}"

    r = session.post(f"{url}", data=objects_to_json(objects, rest_name), params=query_params)

    data = r.json()[rest_name]
    if not as_objects:
        return data

    return json_to_objects(data, obj_type)


def update_objects(
    session: Session,
    url: str,
    objects: List[BaseModel],
    obj_type: Type[BaseModel],
    as_objects=True,
    **query_params,
):

    if not objects:
        return []

    are_same = [o.__class__ == obj_type for o in objects]
    if not all(are_same):
        raise ClientError("Mixed object types")

    rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]

    url = f"{url}/{rest_name}"

    r = session.put(f"{url}", data=objects_to_json(objects, rest_name), params=query_params)

    data = r.json()[rest_name]
    if not as_objects:
        return data

    return json_to_objects(data, obj_type)


def delete_objects(session: Session, url: str, objects: List[BaseModel]):
    if not objects:
        return

    are_same = [o.__class__ == objects[0].__class__ for o in objects[1:]]
    if not all(are_same):
        raise ClientError("Mixed object types")

    obj_type = objects[0].__class__
    rest_name = OBJECT_TYPE_TO_ENDPOINT[obj_type]

    url = f"{url}/{rest_name}"
    data = json.dumps({"source_ids": [obj.id for obj in objects]})

    r = session.delete(url, data=data)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing fitness term definition schema."""

from marshmallow import fields
from marshmallow.validate import OneOf

from ansys.hps.client.common import ObjectSchema

fitness_term_types = ["design_objective", "limit_constraint", "target_constraint"]


class FitnessTermDefinitionSchema(ObjectSchema):
    class Meta(ObjectSchema.Meta):
        pass

    name = fields.String(allow_none=True, metadata={"description": "Name of the fitness term."})
    expression = fields.String(
        allow_none=True,
        metadata={"description": "The Python expression that defines the fitness term."},
    )
    type = fields.String(
        allow_none=True,
        validate=OneOf(fitness_term_types),
        metadata={"description": "Fitness term type."},
    )
    weighting_factor = fields.Float(
        allow_none=True,
        metadata={
            "description": "Relative importance of the fitness "
            "term in comparison to other fitness terms."
        },
    )


class FitnessDefinitionSchema(ObjectSchema):
    class Meta(ObjectSchema.Meta):
        pass

    fitness_term_definitions = fields.Nested(
        FitnessTermDefinitionSchema,
        many=True,
        metadata={"description": "List of :class:`FitnessTermDefinition`."},
    )
    error_fitness = fields.Float(
        metadata={"description": "The default fitness value assigned to failed jobs."}
    )



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# autogenerated code
"""Module providing the operation resource."""
from datetime import datetime
from typing import Any, Dict, List, Union

from marshmallow.utils import missing

from ansys.hps.client.common import Object

from ..schema.operation import OperationSchema


class Operation(Object):
    """Provides the operation resource.

    Parameters
    ----------
    id : str, optional
        Unique ID to access the resource, generated internally by the server on creation.
    name : str, optional
    target : list, optional
    finished : bool, optional
    succeeded : bool, optional
    progress : float, optional
    status : str, optional
    result : dict, optional
    messages : list, optional
    start_time : datetime, optional
    end_time : datetime, optional
    """

    class Meta:
        schema = OperationSchema
        rest_name = "operations"

    def __init__(
        self,
        id: str = missing,
        name: str = missing,
        target: List = missing,
        finished: bool = missing,
        succeeded: bool = missing,
        progress: float = missing,
        status: str = missing,
        result: Dict = missing,
        messages: List = missing,
        start_time: datetime = missing,
        end_time: datetime = missing,
        **kwargs,
    ):
        self.id = id
        self.name = name
        self.target = target
        self.finished = finished
        self.succeeded = succeeded
        self.progress = progress
        self.status = status
        self.result = result
        self.messages = messages
        self.start_time = start_time
        self.end_time = end_time

        self.obj_type = self.__class__.__name__


OperationSchema.Meta.object_class = Operation
